"""venra.query

A module for interacting with the Vespa query api


Related Vespa documentation:
* https://docs.vespa.ai/en/query-api.html
* https://docs.vespa.ai/en/reference/query-api-reference.html


"""

from collections import OrderedDict
import time

import requests

from . import client
from . import config
from . import exceptions


def _api_err_check(response):
    """Error handling for http responses unique to the query api

    Related Vespa documentation:
    * https://docs.vespa.ai/en/reference/query-api-reference.html#http-status-codes
    """
    if response.status_code != 200:
        err = f"unexpected response\n"
        err += f"  response code: {response.status_code}\n"
        err += f"  response  url: {response.url}\n"
        err += f"  response body: {response.text}\n"
        raise exceptions.VespaRequestError(err)


def _vespa_post(qdata):
    """Internal wrapper for search api http call handling"""
    base_uri = f"{config.vespa_host_app}/search/"
    vc = client.get_vespa_client()
    vr = vc.post(f"{base_uri}", json=qdata)
    _api_err_check(vr)
    vr = vr.json()
    return vr


def search(qdata):
    """Run a search with the given parameters and return the complete results

    The return val is the complete query response json loaded as a dictionary
    """
    qstarted = time.time()
    qresults = _vespa_post(qdata)
    qstopped = time.time()
    elapsed = qstopped - qstarted
    qresults["venra"] = {"elapsed_ms": elapsed}
    return qresults


def extract_metrics(qresults):
    """Return key metrics, timings, etc from the search response

    This is an optional convenience function for extracting various key metrics
    from the query response.
    """
    qmet = {}
    qmet["matchcount"] = qresults["root"]["fields"]["totalCount"]
    qmet["totalcount"] = qresults["root"]["coverage"]["documents"]
    qmet["coverage"] = qresults["root"]["coverage"]["coverage"]
    if "timing" in qresults:
        qmet["querytime"] = qresults["timing"]["querytime"]
        qmet["searchtime"] = qresults["timing"]["searchtime"]
        qmet["summaryfetchtime"] = qresults["timing"]["summaryfetchtime"]
    qmet["elapsed_ms"] = round(1000 * qresults["venra"]["elapsed_ms"], 3)
    return qmet


def extract_docs(qresults, fields_only=True):
    """Return an ordered list of doc fields from a search response

    This is an optional convenience function for the common action of extracting
    an ordered list of the docs from the query response.
    """
    docs = []

    # return early if results are empty
    if not qresults or qresults["root"]["fields"]["totalCount"] == 0:
        return docs

    # collect the docs
    for result in qresults["root"]["children"]:

        # skip group node
        if result["id"] == "group:root:0":
            continue

        # add to results
        if fields_only:
            docs.append(result["fields"])
        else:
            docs.append(result)

    return docs


def extract_groups(qresults):
    """Extract the groups from a query response

    This is an optional convenience function. It will build a dictionary which
    then contains each grouplist, which themselves are ordered dicts. This is
    useful for extracting grouping for easy access to build facets etc.

    example:
     select ... | all(group(time.date(some_date)) order(-max(some_date)) max(10) each(output(count())) as(myfacet) )

    return groups:
     groups["myfacet"] = OrderedDict()
     groups["myfacet"]["2022-3-14"] = 108
     groups["myfacet"]["2022-3-13"] = 300
     groups["myfacet"]["2022-3-12"] = 278
     ...
    """
    groups = {}

    # find the root node for the groups
    group_node = None
    for result in qresults["root"]["children"]:
        if result["id"] == "group:root:0":
            group_node = result
            break

    # return early if there is no group info
    if group_node is None:
        return groups

    # collect the results for each grouplist
    for grouplist in group_node["children"]:
        gl = OrderedDict()
        for group in grouplist["children"]:
            gl[group["value"]] = group["fields"]  # ex: item["fields"]["count()"]
        groups[grouplist["label"]] = gl

    return groups


def extract_group_pairs(groups, label, field_name="count()"):
    """From an extracted grouplist, return a list of (value, count) tuples

    This is an optional convenience function for extracting the fundamental
    (key, value) pairs for a grouping/facet.

    Given:
     groups["bazgroup"] = OrderedDict()
     groups["bazgroup"]["bar"]["fields"]["count()"] = 3
     groups["bazgroup"]["foo"]["fields"]["count()"] = 1
     groups["myfacet"] = OrderedDict()
     groups["myfacet"]["2022-3-14"]["fields"]["count()"] = 108
     groups["myfacet"]["2022-3-13"]["fields"]["count()"] = 300
     groups["myfacet"]["2022-3-12"]["fields"]["count()"] = 278

    Calling extract_group_pairs(groups, "myfacet", "count()")
    Returns:
     [("2022-3-14", 108), ("2022-3-13", 300), ("2022-3-12", 278)]
    """
    pairs = []
    for k,f in groups[label].items():
        pairs.append((k, f[field_name]))
    return pairs


def feed(qdata, fields_only=True):
    """Yield all docs for a given query

    example qdata:
    qdata = {}
    qdata["yql"] = f"select * from sources mydocs where userQuery();"
    qdata["hits"] = 3
    qdata["timeout"] = "3300ms"
    ...

    """
    # ensure offset is explicitly set
    if "offset" not in qdata:
        qdata["offset"] = 0

    # initial query
    qresults = search(qdata)
    total_results = qresults["root"]["fields"]["totalCount"]

    # page through results
    doc_count = 0
    while True:

        # extract the docs from the response structure
        docs = extract_docs(qresults, fields_only)

        # yield each doc from the search results
        for d in docs:
            doc_count += 1
            yield d

        # stop if we reached the end of the results
        if doc_count >= total_results:
            break

        # get next page of results
        qdata["offset"] += 1
        qresults = search(qdata)

    return





"""venra.exceptions

A collection of exceptions which encapsulate various Vespa and http request
errors, to assist with application control flow.

"""


class VenraException(Exception):
    """A base exception for unexpected errors"""


class VespaCommunicationError(VenraException):
    """An error connecting to Vespa"""


class VespaRequestError(VenraException):
    """A general error as a result of making a request to Vespa"""


class VespaItemDoesNotExist(VenraException):
    """An error for attempted operation on an item which does not exist"""




"""venra.client

A module for retrieving the internal venra http client. 

This is a thin wrapper around a requests session client. It allows for cross
cutting access throughout the venra package. Additionally it centralizes and
encapsulates some of the common http request boilerplate.
"""

import functools

import requests

from . import config
from . import exceptions


# TODO - switch to a thread-local cache
vclient = None


def reset():
    """Remove initialized client"""
    global vclient
    vclient = None


def get_vespa_client():
    """Get or create a vespa http client and return"""
    global vclient
    if vclient is None:
        vclient = _init_client()
    return vclient


def _init_client():
    """Initialize and return a new requests based client"""

    vc = requests.Session()

    # set a default timeout for all requests
    vc.request = functools.partial(vc.request, timeout=config.timeout_s)

    # wrapp all requests in a Venra specific try/except handler
    vc.request = _enclose_request(vc.request)

    return vc


def _enclose_request(fn):
    """An envelope to handle all http requests in a Venra speciic manner

    The stack of requests + urrllib3 can be a bit noisy with errors and stack
    traces. Venra aims to allow for simple control flow, so this helper
    wraps the call to requests and returns simpler exceptions and messages.

    Additionally, there are some common response patterns across the different
    Vespa apis. So here we can trap certain response codes and raise exception.
    """
    @functools.wraps(fn)
    def wrapper(*args, **kwds):

        # Common exception handling for Requests lib exceptions
        try:
            response = fn(*args, **kwds)

        except requests.exceptions.Timeout:
            err = f"Timeout communicating with Vespa"
            err += f"\n  timeout: {config.timeout_s} seconds"
            err += f"\n      url: {args[1]}"
            raise exceptions.VespaCommunicationError(err)

        except requests.exceptions.RequestException as original_err:
            err = f"Unable to communicate with vespa"
            err += f"\n  upstream err: {original_err}"
            raise exceptions.VespaCommunicationError(err)

        # Common http response handling
        if response.status_code == 500:
            err = f"internal server error {response.status_code} {response.url}"
            raise exceptions.VespaRequestError(err)

        return response

    return wrapper






"""venra.visit

Implementation for the visit subset of the document api


Vespa documentation:
* https://docs.vespa.ai/en/content/visiting.html
* https://docs.vespa.ai/en/reference/document-v1-api-reference.html#visit


"""


from . import client
from . import config
from . import exceptions


def _api_err_check(response):
    """Error handling for http responses unique to the visit api"""

    if response.status_code != 200:
        err = f"unexpected response {response.status_code} {response.url}"
        raise exceptions.VespaRequestError(err)


def _vespa_get(namespace, doctype, selection=None, continuation=None):
    """Internal wrapper for visit api http call handling"""
    base_uri = f"{config.vespa_host_app}/document/v1/{namespace}/{doctype}/docid"

    vc = client.get_vespa_client()

    if not selection and not continuation:
        vr = vc.get(f"{base_uri}")

    elif not selection and continuation:
        vr = vc.get(f"{base_uri}?continuation={continuation}")

    elif selection and not continuation:
        vr = vc.get(f"{base_uri}?selection={selection}")

    elif selection and continuation:
        vr = vc.get(f"{base_uri}?selection={selection}&continuation={continuation}")

    _api_err_check(vr) 
    vr = vr.json()

    return vr


def feed(namespace, doctype, selection=None, fields_only=True):
    """Yield all docs of a given type and selection

    """

    # initial call
    vr = _vespa_get(namespace, doctype, selection, None)

    # page through results
    while True:

        # yield each doc from the results
        for d in vr.get("documents", []):
            if fields_only:
                d = d["fields"]
            yield d

        # check for continuation / more results
        continuation = vr.get("continuation", False)
        if not continuation:
            break

        # retrieve the next set of results
        vr = _vespa_get(namespace, doctype, selection, continuation)

    return




"""venra.document

A module for interacting with the Vespa /document/v1 api.


Related Vespa documentation:
* https://docs.vespa.ai/en/document-v1-api-guide.html
* https://docs.vespa.ai/en/reference/document-v1-api-reference.html
* https://docs.vespa.ai/en/reference/document-json-format.html
* https://docs.vespa.ai/en/reference/document-json-format.html#document-operations
* https://docs.vespa.ai/en/documents.html#fieldsets


"""

import json

from . import client
from . import config
from . import exceptions


def _api_err_check(response):
    """Error handling for http responses unique to the document api

    Related Vespa documentation:
    * https://docs.vespa.ai/en/reference/document-v1-api-reference.html#http-status-codes
    """

    if response.status_code == 404:
        err = f"{response.url}"
        raise exceptions.VespaItemDoesNotExist(err)

    elif response.status_code != 200:
        err = f"unexpected response\n"
        err += f"  response code: {response.status_code}\n"
        err += f"  response  url: {response.url}\n"
        err += f"  response body: {response.text}\n"
        raise exceptions.VespaRequestError(err)


def _vespa_get(namespace, doctype, docid, fieldset="all"):
    """Internal wrapper for document api and http get handling"""
    base_uri = f"{config.vespa_host_app}/document/v1/{namespace}/{doctype}/docid/{docid}?fieldSet=[{fieldset}]"
    vc = client.get_vespa_client()
    vr = vc.get(f"{base_uri}")
    _api_err_check(vr)
    vr = vr.json()
    return vr


def _vespa_post(namespace, doctype, docid, doc):
    """Internal wrapper for document api and http post handling"""
    base_uri = f"{config.vespa_host_app}/document/v1/{namespace}/{doctype}/docid/{docid}"
    vc = client.get_vespa_client()
    vr = vc.post(f"{base_uri}", json=doc)
    _api_err_check(vr)
    return


def _vespa_put(namespace, doctype, docid, update):
    """Internal wrapper for document api and http put handling"""
    base_uri = f"{config.vespa_host_app}/document/v1/{namespace}/{doctype}/docid/{docid}"
    vc = client.get_vespa_client()
    vr = vc.put(f"{base_uri}", json=update)
    _api_err_check(vr)
    return


def _vespa_delete(namespace, doctype, docid):
    """Internal wrapper for document api and http delete handling"""
    base_uri = f"{config.vespa_host_app}/document/v1/{namespace}/{doctype}/docid/{docid}"
    vc = client.get_vespa_client()
    vr = vc.delete(f"{base_uri}")
    _api_err_check(vr)
    return


def get(namespace, doctype, docid, fieldset="all", fields_only=True):
    """Retrieve and return a document"""
    doc = _vespa_get(namespace, doctype, docid, fieldset)
    if fields_only:
        doc = doc["fields"]
    return doc


def put(namespace, doctype, docid, doc):
    """Add or update a whole document"""
    doc_fields = {"fields": doc}
    _vespa_post(namespace, doctype, docid, doc_fields)
    return True


def update(namespace, doctype, docid, operations):
    """Apply a partial update using the assign, add, or remove operator

    operations = [("assign", "field_name_here", "field_value_here"), ...]

    reference:
    * https://docs.vespa.ai/en/reference/document-json-format.html

    Example:
    * https://docs.vespa.ai/en/reference/document-json-format.html#assign-map-field
    {
        "update": "id:mynamespace:food::example",
        "fields": {
            "my_food_scores{Strawberries}": { "assign": "Delicious!" }
        }
    }
    """
    partial_update = {"fields": {}}
    for operation, field_name, field_value in operations:
        partial_update["fields"][field_name] = {operation: field_value}
    _vespa_put(namespace, doctype, docid, partial_update)
    return True


def remove(namespace, doctype, docid):
    """Delete a document"""
    _vespa_delete(namespace, doctype, docid)
    return True





"""venra.system

A module which aggregates access to the wide variety of Vespa apis which 
generally provide system type information.

Venra is generally focused on client interaction with running Vespa systems,
so this module is currently focused on read-only access to system information.


Related Vespa documentation:
* https://docs.vespa.ai/en/reference/state-v1.html
* https://docs.vespa.ai/en/reference/deploy-rest-api-v2.html
* https://docs.vespa.ai/en/reference/config-rest-api-v2.html
* https://docs.vespa.ai/en/reference/application-v2-tenant.html
* https://docs.vespa.ai/en/reference/cluster-v2.html
* https://docs.vespa.ai/en/reference/metrics-v2.html


"""

import requests

from . import client
from . import config
from . import exceptions


def _api_err_check(response):
    """Error handling for http responses unique to the system apis"""

    if response.status_code != 200:
        err = f"unexpected response\n"
        err += f"  response code: {response.status_code}\n"
        err += f"  response  url: {response.url}\n"
        err += f"  response body: {response.text}\n"
        raise exceptions.VespaRequestError(err)


def _vespa_get(full_uri):
    """Internal wrapper for system api http call handling"""
    vc = client.get_vespa_client()
    vr = vc.get(f"{full_uri}")
    _api_err_check(vr)
    vr = vr.json()
    return vr


def version_app():
    """Return the version string from the app server

    example:
    $ curl -s http://127.0.0.1:8080/state/v1/version | jq .
    {
        "version": "8.13.21"
    }
    """
    vr = _vespa_get(f"{config.vespa_host_app}/state/v1/version")
    return vr["version"]


def version_cfg():
    """Return the version string from the app server

    example:
    $ curl -s http://127.0.0.1:19071/state/v1/version | jq .
    {
        "version": "8.13.21"
    }
    """
    vr = _vespa_get(f"{config.vespa_host_cfg}/state/v1/version")
    return vr["version"]


